%# Copyright (C) 2009-2012, Sebastian Schoeps <schoeps@math.uni-wuppertal.de>
%# Copyright (C) 2013, Roberto Porcu' <roberto.porcu@polimi.it>
%# OdePkg - A package for solving ordinary differential equations and more
%#
%# This program is free software; you can redistribute it and/or modify
%# it under the terms of the GNU General Public License as published by
%# the Free Software Foundation; either version 2 of the License, or
%# (at your option) any later version.
%#
%# This program is distributed in the hope that it will be useful,
%# but WITHOUT ANY WARRANTY; without even the implied warranty of
%# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%# GNU General Public License for more details.
%#
%# You should have received a copy of the GNU General Public License
%# along with this program; If not, see <http://www.gnu.org/licenses/>.

%# -*- texinfo -*-
%# @deftypefn {Command} {[@var{J}] =} approx_Jacobian (@var{@@fun}, @var{t}, @var{x})
%# 
%# This function file can be used to approximate the Jacobian of a function handle for ODE solver @command{odebwe}.
%# 
%# If no explicit expression for the Jacobian is given to @command{odebwe}, it calls this function to approximate it.
%# 
%# The first input argument must be a function_handle or an inline function and must define the ODE system.
%# 
%# The second and third input argument are just the time and the point at which the Jacobian will be evaluated.
%#
%# @seealso{odepkg}
%# @end deftypefn

function df = approx_Jacobian (vfun, vthetime, vtheinput);

  vlen   = length(vtheinput);

  vsigma = sqrt(eps);

  vfun0 = feval(vfun,vthetime,vtheinput);

  df = zeros(vlen,vlen);

  for j = 1:vlen

     vbuffer = vtheinput(j);

     if (vbuffer==0)
        vh = vsigma;
     elseif (abs(vbuffer)>1)
        vh = vsigma*vbuffer;
     else
        vh = sign(vbuffer)*vsigma;
     end

     vtheinput(j) = vbuffer + vh;

     df(:,j) = (feval(vfun,vthetime,vtheinput) - vfun0) / vh;   

     vtheinput(j) = vbuffer;
  end
end

%!function [ydot] = fpol (vt, vy) %# The Van der Pol
%!  ydot = [vy(2); (1 - vy(1)^2) * vy(2) - vy(1)];
%!endfunction
%!
%! %# Turn off output of warning messages for all tests, turn them on
%! %# again if the last test is called
%!test
%!  warning ('off', 'OdePkg:InvalidArgument');
%!  res = approx_Jacobian (@fpol, 0, [2 0]);
%!
%!  warning ('on', 'OdePkg:InvalidArgument');

%# Local Variables: ***
%# mode: octave ***
%# End: ***
